# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import abc
from dataclasses import dataclass
from typing import Optional, List, Dict, Any, Iterable

from odxtools.utils import read_description_from_odx, make_ref

from .dataobjectproperty import DopBase
from .globals import logger


class TableBase(abc.ABC):
    """ Base class for all Tables."""

    def __init__(self, id: str, short_name: str, long_name=None):
        self.id = id
        self.short_name = short_name
        self.long_name = long_name


@dataclass
class TableRow:
    """This class represents a TABLE-ROW."""

    id: str
    short_name: str
    long_name: str
    key: int
    structure_ref: Optional[str] = None
    dop_ref: Optional[str] = None
    description: Optional[str] = None
    semantic: Optional[str] = None

    def __post_init__(self):
        self._structure: Optional[DopBase] = None
        self._dop: Optional[DopBase] = None

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        if self.structure_ref is not None:
            self._structure = id_lookup.get(self.structure_ref)
            if self._structure is None:
                logger.warning(
                    f"STRUCTURE-REF '{self.structure_ref}' could not be resolved in TABLE-ROW."
                )
        if self.dop_ref is not None:
            self._dop = id_lookup.get(self.dop_ref)
            if self._dop is None:
                logger.warning(
                    f"DATA-OBJECT-PROP-REF '{self.dop_ref}' could not be resolved in TABLE-ROW."
                )

    @property
    def structure(self) -> Optional[DopBase]:
        """The structure associated with this table row."""
        return self._structure

    @property
    def dop(self) -> Optional[DopBase]:
        """The dop object resolved by dop_ref."""
        return self._dop

    def __repr__(self) -> str:
        return (
            f"TableRow('{self.short_name}', "
            + ", ".join(
                [
                    f"key='{self.key}'",
                    f"structure_ref='{self.structure_ref}'",
                    f"dop_ref='{self.dop_ref}'",
                ]
            )
            + ")"
        )

class Table(TableBase):
    """This class represents a TABLE."""

    def __init__(
        self,
        id: str,
        short_name: str,
        table_rows: List[TableRow],
        table_row_refs: Optional[List[str]] = None,
        long_name: Optional[str] = None,
        key_dop_ref: Optional[str] = None,
        description: Optional[str] = None,
        semantic: Optional[str] = None,
    ):
        super().__init__(
            id=id, short_name=short_name, long_name=long_name
        )
        self._local_table_rows = table_rows
        self._ref_table_rows: List[TableRow] = []
        self._table_row_refs = table_row_refs or []
        self.key_dop_ref = key_dop_ref
        self._key_dop = None
        self.description = description
        self.semantic = semantic

    @property
    def key_dop(self) -> Optional[DopBase]:
        """The key data object property associated with this table."""
        return self._key_dop

    @property
    def table_rows(self) -> Iterable[TableRow]:
        """The table rows (both local and referenced) in this table."""
        return self._local_table_rows + self._ref_table_rows

    def _build_id_lookup(self):
        id_lookup = {}
        id_lookup.update({table_row.id: table_row for table_row in self.table_rows})
        return id_lookup

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        if self.key_dop_ref is not None:
            self._key_dop = id_lookup.get(self.key_dop_ref)
            if self._key_dop is None:
                logger.warning(
                    f"KEY-DOP-REF '{self.key_dop_ref!r}' could not be resolved in TABLE."
                )
        for table_row in self._local_table_rows:
            table_row._resolve_references(id_lookup)

        self._ref_table_rows = []
        for ref in self._table_row_refs:
            tr = id_lookup.get(ref)
            if isinstance(tr, TableRow):
                self._ref_table_rows.append(tr)

    def __repr__(self) -> str:
        return (
            f"Table('{self.short_name}', "
            + ", ".join(
                [f"table_rows='{self.table_rows}'", f"key_dop_ref='{self.key_dop_ref}'"]
            )
            + ")"
        )


def _get_common_props(et_element):
    description = read_description_from_odx(et_element.find("DESC"))
    return dict(
        id=et_element.get("ID"),
        short_name=et_element.findtext("SHORT-NAME"),
        long_name=et_element.findtext("LONG-NAME"),
        semantic=et_element.get("SEMANTIC"),
        description=description,
    )


def read_table_row_from_odx(et_element):
    """Reads a TABLE-ROW."""
    key = et_element.find("KEY").text
    structure_ref = None
    if et_element.find("STRUCTURE-REF") is not None:
        structure_ref = make_ref(et_element.find("STRUCTURE-REF"))
    dop_ref = None
    if et_element.find("DATA-OBJECT-PROP-REF") is not None:
        dop_ref = make_ref(et_element.find("DATA-OBJECT-PROP-REF"))

    table_row = TableRow(
        key=key,
        structure_ref=structure_ref,
        dop_ref=dop_ref,
        **_get_common_props(et_element)
    )

    return table_row


def read_table_from_odx(et_element):
    """Reads a TABLE."""
    short_name = et_element.find("SHORT-NAME").text
    key_dop_ref = None
    if et_element.find("KEY-DOP-REF") is not None:
        key_dop_ref = make_ref(et_element.find("KEY-DOP-REF"))
    logger.debug("Parsing TABLE " + short_name)

    table_rows = [
        read_table_row_from_odx(el) for el in et_element.iterfind("TABLE-ROW")
    ]

    table_row_refs = [
        make_ref(el) for el in et_element.iterfind("TABLE-ROW-REF")
    ]

    table = Table(
        key_dop_ref=key_dop_ref,
        table_rows=table_rows,
        table_row_refs=table_row_refs,
        **_get_common_props(et_element)
    )

    return table



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass, field
from itertools import chain
from typing import Optional

from .dataobjectproperty import (
    DataObjectProperty,
    DtcDop,
    read_data_object_property_from_odx,
)
from .endofpdufield import EndOfPduField, read_end_of_pdu_field_from_odx
from .envdata import read_env_data_from_odx, EnvironmentData
from .envdatadesc import read_env_data_desc_from_odx, EnvironmentDataDescription
from .globals import logger
from .multiplexer import Multiplexer, read_mux_from_odx
from .nameditemlist import NamedItemList
from .structures import Structure, read_structure_from_odx
from .table import read_table_from_odx, Table
from .units import read_unit_spec_from_odx, UnitSpec


def _construct_named_item_list(iterable):
    return NamedItemList(lambda x: x.short_name, iterable)


@dataclass()
class DiagDataDictionarySpec:
    dtc_dops: NamedItemList[DtcDop] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    data_object_props: NamedItemList[DataObjectProperty] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    structures: NamedItemList[Structure] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    end_of_pdu_fields: NamedItemList[EndOfPduField] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    tables: NamedItemList[Table] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    env_data_descs: NamedItemList[EnvironmentDataDescription] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    env_datas: NamedItemList[EnvironmentData] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    muxs: NamedItemList[Multiplexer] = field(
        default_factory=lambda: _construct_named_item_list([])
    )
    unit_spec: Optional[UnitSpec] = None

    def __post_init__(self):
        self._all_data_object_properties = _construct_named_item_list(
            chain(
                self.data_object_props,
                self.structures,
                self.end_of_pdu_fields,
                self.dtc_dops,
                self.tables,
            )
        )

        # The attributes are already initialized as (most likely normal) lists.
        # Convert them to NamedItemLists if necessary
        if not isinstance(self.dtc_dops, NamedItemList):
            self.dtc_dops = _construct_named_item_list(self.dtc_dops)

        if not isinstance(self.data_object_props, NamedItemList):
            self.data_object_props = _construct_named_item_list(self.data_object_props)

        if not isinstance(self.structures, NamedItemList):
            self.structures = _construct_named_item_list(self.structures)

        if not isinstance(self.end_of_pdu_fields, NamedItemList):
            self.end_of_pdu_fields = _construct_named_item_list(self.end_of_pdu_fields)

        if not isinstance(self.tables, NamedItemList):
            self.tables = _construct_named_item_list(self.tables)

        if not isinstance(self.env_data_descs, NamedItemList):
            self.env_data_descs = _construct_named_item_list(self.env_data_descs)

        if not isinstance(self.env_datas, NamedItemList):
            self.env_datas = _construct_named_item_list(self.env_datas)

        if not isinstance(self.muxs, NamedItemList):
            self.muxs = _construct_named_item_list(self.muxs)

    def _build_id_lookup(self):
        id_lookup = {}
        for obj in chain(
            self.data_object_props, self.structures, self.end_of_pdu_fields, self.tables
        ):
            id_lookup[obj.id] = obj

        for table in self.tables:
            id_lookup.update(table._build_id_lookup())

        for env_data_desc in self.env_data_descs:
            id_lookup.update(env_data_desc._build_id_lookup())

        for env_data in self.env_datas:
            id_lookup.update(env_data._build_id_lookup())

        for mux in self.muxs:
            id_lookup.update(mux._build_id_lookup())

        for obj in self.dtc_dops:
            id_lookup.update(obj._build_id_lookup())

        if self.unit_spec:
            id_lookup.update(self.unit_spec._build_id_lookup())

        return id_lookup

    def _resolve_references(self, parent_dl, id_lookup: dict):
        for dop in chain(
            self.dtc_dops,
            self.data_object_props,
            self.tables,
        ):
            dop._resolve_references(id_lookup)

        for struct in chain(self.structures, self.end_of_pdu_fields):
            struct._resolve_references(parent_dl, id_lookup)

        for env_data in chain(
            self.env_datas,
        ):
            env_data._resolve_references(parent_dl, id_lookup)

        for mux in chain(
            self.muxs,
        ):
            mux._resolve_references(id_lookup)

        if self.unit_spec:
            self.unit_spec._resolve_references(id_lookup)

    @property
    def all_data_object_properties(self):
        return self._all_data_object_properties


def read_diag_data_dictionary_spec_from_odx(et_element):
    # Parse DOP-BASEs
    data_object_props = [
        read_data_object_property_from_odx(dop_element)
        for dop_element in et_element.iterfind("DATA-OBJECT-PROPS/DATA-OBJECT-PROP")
    ]

    structures = [
        read_structure_from_odx(structure_element)
        for structure_element in et_element.iterfind("STRUCTURES/STRUCTURE")
    ]

    end_of_pdu_fields = [
        read_end_of_pdu_field_from_odx(eopf_element)
        for eopf_element in et_element.iterfind("END-OF-PDU-FIELDS/END-OF-PDU-FIELD")
    ]

    dtc_dops = [
        read_data_object_property_from_odx(dop_element)
        for dop_element in et_element.iterfind("DTC-DOPS/DTC-DOP")
    ]

    tables = [
        read_table_from_odx(table_element)
        for table_element in et_element.iterfind("TABLES/TABLE")
    ]

    env_data_descs = [
        read_env_data_desc_from_odx(env_data_desc_element)
        for env_data_desc_element in et_element.iterfind("ENV-DATA-DESCS/ENV-DATA-DESC")
    ]

    env_data_elements = chain(
        et_element.iterfind("ENV-DATAS/ENV-DATA"),
        # ODX 2.0.0 says ENV-DATA-DESC could contain a list of ENV-DATAS
        et_element.iterfind("ENV-DATA-DESCS/ENV-DATA-DESC/ENV-DATAS/ENV-DATA"),
    )
    env_datas = [
        read_env_data_from_odx(env_data_element)
        for env_data_element in env_data_elements
    ]

    muxs = [
        read_mux_from_odx(mux_element)
        for mux_element in et_element.iterfind("MUXS/MUX")
    ]

    if et_element.find("UNIT-SPEC") is not None:
        unit_spec = read_unit_spec_from_odx(et_element.find("UNIT-SPEC"))
    else:
        unit_spec = None

    # TODO: Parse different specs.. Which of them are needed?
    for (path, name) in [
        ("STATIC-FIELDS", "static fields"),
        ("DYNAMIC-LENGTH-FIELDS/DYNAMIC-LENGTH-FIELD", "dynamic length fields"),
        (
            "DYNAMIC-ENDMARKER-FIELDS/DYNAMIC-ENDMARKER-FIELD",
            "dynamic endmarker fields",
        ),
    ]:
        num = len(list(et_element.iterfind(path)))
        if num > 0:
            logger.info(f"Not implemented: Did not parse {num} {name}.")

    return DiagDataDictionarySpec(
        data_object_props=data_object_props,
        structures=structures,
        end_of_pdu_fields=end_of_pdu_fields,
        dtc_dops=dtc_dops,
        unit_spec=unit_spec,
        tables=tables,
        env_data_descs=env_data_descs,
        env_datas=env_datas,
        muxs=muxs,
    )



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from .parameterbase import Parameter


class TableStructParameter(Parameter):
    def __init__(self,
                 short_name,
                 table_key_ref=None,
                 table_key_snref=None,
                 long_name=None,
                 byte_position=None,
                 bit_position=0,
                 semantic=None,
                 description=None):
        super().__init__(
            short_name=short_name,
            long_name=long_name,
            byte_position=byte_position,
            bit_position=bit_position,
            parameter_type="TABLE-STRUCT",
            semantic=semantic,
            description=description
        )
        if table_key_ref:
            self.table_key_ref = table_key_ref
        elif table_key_snref:
            self.table_key_snref = table_key_snref
        else:
            raise ValueError(
                "Either table_key_ref or table_key_snref must be defined.")

    def is_required(self):
        raise NotImplementedError(
            "TableStructParameter.is_required is not implemented yet.")

    def is_optional(self):
        raise NotImplementedError(
            "TableStructParameter.is_optional is not implemented yet.")

    def get_coded_value(self):
        raise NotImplementedError(
            "Encoding a TableStructParameter is not implemented yet.")

    def get_coded_value_as_bytes(self):
        raise NotImplementedError(
            "Encoding a TableStructParameter is not implemented yet.")

    def decode_from_pdu(self, coded_message, default_byte_position=None):
        raise NotImplementedError(
            "Decoding a TableStructParameter is not implemented yet.")



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import Iterable, Union

from .dataobjectproperty import DataObjectProperty
from .endofpdufield import EndOfPduField
from .parameters import CodedConstParameter
from .parameters import MatchingRequestParameter
from .parameters import ReservedParameter
from .parameters import Parameter
from .parameters import ParameterWithDOP
from .endofpdufield import EndOfPduField
from .compumethods import Limit
from .compumethods import IntervalType
from .compumethods import IdenticalCompuMethod
from .compumethods import LinearCompuMethod
from .compumethods import TexttableCompuMethod

import re

def parameter_info(param_list : Iterable[Union[Parameter, EndOfPduField]]) -> str:
    result = ""
    for param in param_list:
        if isinstance(param, CodedConstParameter):
            result += f"{param.short_name} : const = {param._coded_value_str}\n"
            continue
        elif isinstance(param, MatchingRequestParameter):
            result += f"{param.short_name} : <matches request>\n"
            continue
        elif isinstance(param, ReservedParameter):
            result += f"{param.short_name} : <reserved>\n"
            continue
        elif not isinstance(param, ParameterWithDOP):
            result += f"{param.short_name} : <unhandled parameter type>\n"
            continue

        dop = param.dop

        if isinstance(dop, EndOfPduField):
            result += f"{param.short_name} : <optional> list({{\n"
            tmp = parameter_info(dop.structure.parameters).strip()
            tmp = re.sub("^", "  ", tmp)
            result += tmp + "\n"
            result += f"}})\n"
            continue

        result += f"{param.short_name}"

        if dop is None:
            result += ": <no DOP>\n"
            continue
        elif not isinstance(dop, DataObjectProperty):
            result += ": <unhandled DOP>\n"
            continue

        if (cm := dop.compu_method) is None:
            result += ": <no compu method>\n"
            continue

        if isinstance(cm, TexttableCompuMethod):
            result += f": enum; choices:\n"
            for scale in cm.internal_to_phys:
                result += f"  '{scale.compu_const}'\n"

        elif isinstance(cm, IdenticalCompuMethod):
            bdt = dop.physical_type.base_data_type.name
            if bdt in ("A_UTF8STRING", "A_UNICODE2STRING", "A_ASCIISTRING"):
                result += f": str"
            elif bdt in ("A_BYTEFIELD"):
                result += f": bytes"
            elif bdt.startswith("A_FLOAT"):
                result += f": float"
            elif bdt.startswith("A_UINT"):
                result += f": uint"
            elif bdt.startswith("A_INT"):
                result += f": int"
            else:
                result += f": <unknown type>"

            if dop.bit_length is not None:
                result += f"{dop.bit_length}\n"
            else:
                result += "\n"

        elif isinstance(cm, LinearCompuMethod):
            result += f": float\n"
            ll = cm.physical_lower_limit
            ul = cm.physical_upper_limit
            result += \
                f" range: " \
                f"{'[' if ll.interval_type == IntervalType.CLOSED else '('}" \
                f"{ll.value}, " \
                f"{ul.value}" \
                f"{']' if ul.interval_type == IntervalType.CLOSED else ')'}\n"

    return result



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import math
from typing import List, Dict, Iterable, OrderedDict, Tuple, Union
import warnings

from odxtools.parameters.tablekeyparameter import TableKeyParameter
from odxtools.parameters.lengthkeyparameter import LengthKeyParameter

from .dataobjectproperty import DataObjectProperty, DopBase
from .decodestate import DecodeState, ParameterValuePair
from .encodestate import EncodeState
from .exceptions import DecodeError, EncodeError, OdxWarning
from .globals import logger
from .nameditemlist import NamedItemList
from .parameters import Parameter, ParameterWithDOP, read_parameter_from_odx
from .parameters import CodedConstParameter, MatchingRequestParameter, ValueParameter
from .utils import read_description_from_odx


class BasicStructure(DopBase):
    def __init__(self,
                 id,
                 short_name,
                 parameters: Iterable[Union[Parameter, "EndOfPduField"]], # type: ignore
                 long_name=None,
                 byte_size=None,
                 description=None):
        super().__init__(id, short_name, long_name=long_name, description=description)
        self.parameters : NamedItemList[Union[Parameter, "EndOfPduField"]] = NamedItemList(lambda par: par.short_name, parameters) # type: ignore

        self._byte_size = byte_size

    @property
    def bit_length(self):
        if self._byte_size:
            return 8 * self._byte_size
        elif all(p.bit_length is not None for p in self.parameters):
            offset = 0
            length = 0
            for param in self.parameters:
                if isinstance(param, ValueParameter) and hasattr(param.dop, 'min_number_of_items'):
                    # The param repeats itself, making bit_length calculation invalid
                    # Temporary workaround
                    # Can not import EndOfPduField to check on its type due to circular dependency
                    return None
                if param.byte_position is not None:
                    offset = param.byte_position * 8 + param.bit_position

                offset += param.bit_length
                length = max(length, offset)
   
            return length
        else:
            return None

    def coded_const_prefix(self, request_prefix: Union[bytes, bytearray] = bytes()):
        prefix = bytearray()
        encode_state = EncodeState(
            prefix, parameter_values={}, triggering_request=request_prefix)
        for p in self.parameters:
            if isinstance(p, CodedConstParameter) and p.bit_length % 8 == 0:
                prefix = p.encode_into_pdu(encode_state)
                encode_state = EncodeState(prefix, *encode_state[1:])
            elif isinstance(p, MatchingRequestParameter):
                prefix = p.encode_into_pdu(encode_state)
                encode_state = EncodeState(prefix, *encode_state[1:])
            else:
                break
        return prefix

    def get_required_parameters(self) -> List[Parameter]:
        """Return the list of parameters which are required for
        encoding the structure."""
        return [p for p in self.parameters if p.is_required()]

    def get_free_parameters(self) -> List[Union[Parameter, "EndOfPduField"]]: # type: ignore
        """Return the list of parameters which can be freely specified by
        the user when encoding the structure.

        This means all required parameters plus the parameters that
        can be omitted minus those which are implicitly specified by
        the corresponding request (in the case of responses).

        """
        from .endofpdufield import EndOfPduField

        result : List[Union[Parameter, EndOfPduField]] = []
        for param in self.parameters:
            if isinstance(param, EndOfPduField):
                result.append(param)
                continue
            elif not param.is_required():
                continue
            # The user cannot specify MatchingRequestParameters freely!
            elif isinstance(param, MatchingRequestParameter):
                continue
            result.append(param)

        return result

    def free_parameters_info(self) -> str:
        """Return a human readable description of the structure's
        free parameters.
        """
        from .parameter_info import parameter_info

        return parameter_info(self.get_free_parameters())

    def convert_physical_to_internal(self,
                                     param_values: dict,
                                     triggering_coded_request,
                                     is_end_of_pdu=True):
        logger.debug(f"{self.short_name} encode RPC"
                     f" with params={param_values}")

        coded_rpc = bytearray()
        encode_state = EncodeState(coded_rpc,
                                   dict(param_values),
                                   triggering_request=triggering_coded_request,
                                   is_end_of_pdu=False)

        length_encodings: List[Tuple[LengthKeyParameter, EncodeState]] = []
        for param in self.parameters:
            if param == self.parameters[-1]:
                # The last parameter is at the end of the PDU if the structure itself is at the end of the PDU
                encode_state = encode_state._replace(
                    is_end_of_pdu=is_end_of_pdu)

            implicit_length_encoding = isinstance(param, LengthKeyParameter) and param.short_name not in param_values
            if implicit_length_encoding:
                # Mark this parameter since we need to re-encode it later on
                length_encodings.append((param, encode_state))
                # Give it a default value for now
                encode_state.parameter_values[param.short_name] = 0

            coded_rpc = param.encode_into_pdu(encode_state)
            encode_state = encode_state._replace(coded_message=coded_rpc)

            if implicit_length_encoding:
                # Undo length_keys changes
                encode_state.length_keys.pop(param.id)

        if self._byte_size is not None and len(coded_rpc) < self._byte_size:
            # Padding bytes needed
            coded_rpc = coded_rpc.ljust(self._byte_size, b'\0')

        for (param, encode_state) in length_encodings:
            # Same as previous, but all bytes as 0
            param_value = encode_state.length_keys[param.id]
            state = encode_state._replace(
                coded_message=bytearray(len(encode_state.coded_message)),
                parameter_values={param.short_name: param_value},
            )
            # Encode the length into the zeros coded message
            param_bytes = param.encode_into_pdu(state)
            # Bits that changed value needs to be updated in coded_rpc
            for i, b in enumerate(param_bytes):
                coded_rpc[i] |= b

        # Assert that length is as expected
        self._validate_coded_rpc(coded_rpc)

        return bytearray(coded_rpc)

    def _validate_coded_rpc(
            self,
            coded_rpc: bytearray):
        
        if self._byte_size is not None:
            # We definitely broke something if we didn't respect the explicit byte_size
            assert len(coded_rpc) == self._byte_size, self._get_encode_error_str('was', coded_rpc, self._byte_size * 8)
            # forget what we calcualted, ODX defined an explicit byte_size
            # if bit_length was less than _byte_size then we have padding or gaps between parameters
            # if bit_length was more than _byte_size then we over calculated (could happen with overlapping parameters)
            return

        bit_length = self.bit_length

        if bit_length is None:
            # Nothing to check
            return

        if bit_length % 8 != 0:
            warnings.warn(
                f"Structure {self.short_name} length {bit_length} is not divisible by 8, i.e. is not a full sequence of bytes.",
                OdxWarning)
            # Round up so not to trigger double alarms
            bit_length = math.ceil(bit_length / 8) * 8

        if len(coded_rpc) * 8 != bit_length:
            # We may have broke something
            # but it could be that bit_length was mis calculated and not the actual bytes are wrong
            # Could happen with overlapping parameters and parameters with gaps
            warnings.warn(
                self._get_encode_error_str('may have been', coded_rpc, bit_length),
                OdxWarning)

    def _get_encode_error_str(self,
            verb: str,
            coded_rpc: bytearray,
            bit_length: int):

        return str(f"Structure {self.short_name} {verb} encoded uncorrectly:" +
                f" actual length is {len(coded_rpc)}," +
                f" computed byte length is {bit_length // 8}," +
                f" computed_rpc is {coded_rpc.hex()}\n" +
                '\n'.join(self.__message_format_lines()))


    def convert_physical_to_bytes(self, param_values: dict, encode_state: EncodeState, bit_position=0):
        if bit_position != 0:
            raise EncodeError("Structures must be aligned, i.e. bit_position=0, but "
                              f"{self.short_name} was passed the bit position {bit_position}")
        return self.convert_physical_to_internal(param_values,
                                                 triggering_coded_request=encode_state.triggering_request,
                                                 is_end_of_pdu=encode_state.is_end_of_pdu
                                                 )

    def convert_bytes_to_physical(self, decode_state: DecodeState, bit_position=0):
        if bit_position != 0:
            raise DecodeError("Structures must be aligned, i.e. bit_position=0, but "
                              f"{self.short_name} was passed the bit position {bit_position}")
        byte_code = decode_state.coded_message[decode_state.next_byte_position:]
        inner_decode_state = DecodeState(coded_message=byte_code,
                                         parameter_value_pairs=[],
                                         next_byte_position=0)

        for parameter in self.parameters:
            value, next_byte_position = parameter.decode_from_pdu(
                inner_decode_state)

            inner_decode_state.parameter_value_pairs.append(
                ParameterValuePair(parameter, value))
            inner_decode_state = DecodeState(coded_message=byte_code,
                                             parameter_value_pairs=inner_decode_state.parameter_value_pairs,
                                             next_byte_position=max(inner_decode_state.next_byte_position,
                                                                    next_byte_position))
        # Construct the param dict.
        # TODO: Wouldn't it be prettier if we kept the information of each parameter
        #       instead of just using the short_name as the key and "forgetting" everything else?
        param_dict = OrderedDict((pv.parameter.short_name, pv.value)
                                 for pv in inner_decode_state.parameter_value_pairs)

        return param_dict, decode_state.next_byte_position + inner_decode_state.next_byte_position

    def encode(self, coded_request=None, **params) -> bytearray:
        """
        Composes an UDS message as bytes for this service.
        Parameters:
        ----------
        coded_request: bytes
            coded request (only needed when encoding a response)
        params: dict
            Parameters of the RPC as mapping from SHORT-NAME of the parameter to the value
        """
        return self.convert_physical_to_internal(params,
                                                 triggering_coded_request=coded_request,
                                                 is_end_of_pdu=True)

    def decode(self, message: Union[bytes, bytearray]):
        # dummy decode state to be passed to convert_bytes_to_physical
        decode_state = DecodeState(coded_message=message,
                                   parameter_value_pairs=[],
                                   next_byte_position=0)
        param_values, next_byte_position = self.convert_bytes_to_physical(
            decode_state)
        if len(message) != next_byte_position:
            raise DecodeError(
                f"The message {message.hex()} is longer than could be parsed. Expected {next_byte_position} but got {len(message)}.")
        return param_values

    def parameter_dict(self):
        """
        Returns a dict with parameter short names as keys.
        The values are parameters for simple types or a nested dict for structures.
        """
        params = self.parameters
        assert all(not isinstance(p, ParameterWithDOP) or isinstance(
            p.dop, DataObjectProperty) or isinstance(p.dop, Structure) for p in self.parameters)
        param_dict: Dict[str, Union[Parameter, Dict[str, Parameter]]] = {
            p.short_name: p for p in params
            if not isinstance(p, ParameterWithDOP) or not isinstance(p.dop, Structure)
        }
        param_dict.update({
            struct_param.short_name: struct_param.dop.parameter_dict()
            for struct_param in params if isinstance(struct_param, ParameterWithDOP) and isinstance(struct_param.dop, BasicStructure)
        })
        return param_dict

    def _resolve_references(self, parent_dl, id_lookup):
        """Recursively resolve any references (odxlinks or sn-refs)
        """
        for p in self.parameters:
            if isinstance(p, ParameterWithDOP):
                p.resolve_references(parent_dl, id_lookup)
            if isinstance(p, TableKeyParameter):
                p.resolve_references(parent_dl, id_lookup)

    def __message_format_lines(self, allow_unknown_lengths=False):
        # sort parameters
        sorted_params: list = list(self.parameters)  # copy list
        if all(p.byte_position is not None for p in self.parameters):
            sorted_params.sort(key=lambda p: (
                p.byte_position, 8 - p.bit_position))

        # replace structure parameters by their sub parameters
        params = []
        for p in sorted_params:
            if isinstance(p, ValueParameter) and isinstance(p.dop, BasicStructure):
                params += p.dop.parameters
            else:
                params.append(p)

        i = 0
        byte = 0
        # needs to be one larger than the maximum digit length of a byte number
        indent_for_byte_numbering = 5 * " "
        formatted_lines = [indent_for_byte_numbering +
                           "".join(f"   {7-bit}  " for bit in range(8))]

        breakpoint = 0  # absolute bit position where the next parameter starts

        divide_string = indent_for_byte_numbering + 8 * "+-----" + "+"

        error = False
        next_line = ""
        while i <= len(params) and not error:  # For each byte
            if 8 * byte == breakpoint and i == len(params):
                # If we have formatted the last parameter, we're done.
                break

            formatted_lines.append(f"{divide_string}")
            if breakpoint // 8 - byte > 5:
                curr_param = params[i-1].short_name
                formatted_lines.append(
                    indent_for_byte_numbering + f"  ... {breakpoint // 8 - byte} bytes belonging to {curr_param} ... ")
                byte += breakpoint // 8 - byte
                continue

            next_line = f"{(len(indent_for_byte_numbering) - 1 - len(str(byte))) * ' '}{byte} "

            for bit in range(8):
                assert 8 * byte + bit <= breakpoint

                if 8 * byte + bit == breakpoint:
                    # END-OF-PDU fields do not exhibit a fixed bit
                    # length, so they need special treatment here
                    dct = None
                    if hasattr(params[i], 'dop'):
                        dop = params[i].dop
                        if hasattr(dop, 'diag_coded_type'):
                            dct = dop.diag_coded_type

                    if dct is not None and dct.dct_type == 'MIN-MAX-LENGTH-TYPE':
                        name = params[i].short_name + " ("
                        if dct.termination == "END-OF-PDU":
                            name += "End of PDU, "
                        name += f"{dct.min_length}..{dct.max_length} bytes"
                        name += ")"
                        next_line += "| " + name

                        i += 1

                        # adding 8 is is a bit hacky here, but hey, it
                        # works ...
                        breakpoint += 8

                        break

                    elif not params[i].bit_length and not allow_unknown_lengths:
                        # The bit length is not set for the current
                        # parameter, i.e. it was either not specified
                        # or the parameter is of variable length and
                        # has an type which is not handled above. In
                        # this case, stop trying.
                        error = True
                        break
                    else:
                        breakpoint += params[i].bit_length or (
                            allow_unknown_lengths and 8)
                        name = params[i].short_name + \
                            f" ({params[i].bit_length or 'Unknown'} bits)"
                        next_line += "| " + name

                    i += 1

                    if byte == breakpoint // 8:
                        char_pos = bit * 6 + 2 + len(name)
                        width_of_line = (breakpoint % 8) * 6
                        if char_pos < width_of_line:
                            next_line += " " * \
                                (width_of_line - char_pos) + "|"
                        # start next line (belongs to same byte)
                        formatted_lines.append(next_line)
                        # fill next line with white spaces upto the bit where next parameter starts
                        next_line = indent_for_byte_numbering + \
                            (bit + 1) * 6 * " "
                    else:
                        char_pos = 2 + bit * 6 + len(name)
                        width_of_line = 8 * 6
                        if char_pos < width_of_line:
                            next_line += " " * \
                                (width_of_line - char_pos) + "|"
                        break
                else:
                    if bit == 0:
                        next_line += "|" + 5 * " "
                    elif bit == 7:
                        next_line += 6 * " " + "|"
                    else:
                        next_line += 6 * " "

            formatted_lines.append(next_line)
            next_line = ""

            byte += 1

        if not error:
            formatted_lines.append(divide_string)
            return formatted_lines
        else:
            return None

    def print_message_format(self, indent: int = 5, allow_unknown_lengths=False):
        """
        Print a description of the message format to `stdout`.
        """

        message_as_lines = self.__message_format_lines(
            allow_unknown_lengths=allow_unknown_lengths)
        if message_as_lines is not None:
            print(f"{indent * ' '}" +
                  f"\n{indent * ' '}".join(message_as_lines))
        else:
            print("Sorry, couldn't pretty print message layout. :(")
        for p in self.parameters:
            print(indent * ' ' +
                  str(p).replace("\n", f"\n{indent * ' '}"))


class Structure(BasicStructure):
    def __init__(self, id, short_name, parameters, long_name=None, byte_size=None, description=None):
        super().__init__(id, short_name, parameters,
                         long_name=long_name, description=description)

        self.parameters = parameters
        self.basic_structure = BasicStructure(id, short_name, parameters,
                                              long_name=long_name, byte_size=byte_size, description=description)

    def __repr__(self) -> str:
        return f"Structure('{self.short_name}', byte_size={self._byte_size})"

    def __str__(self) -> str:
        params = "[\n" + "\n".join([" " + str(p).replace("\n", "\n ")
                                   for p in self.parameters]) + "\n]"

        return \
            f"Structure '{self.short_name}': " + \
            f"Byte size={self._byte_size}, " + \
            f"Parameters={params}"


class Request(BasicStructure):
    def __init__(self, id, short_name, parameters, long_name=None, description=None):
        super().__init__(id, short_name, parameters,
                         long_name=long_name, description=description)

    def __repr__(self) -> str:
        return f"Request('{self.short_name}')"

    def __str__(self) -> str:
        return f"Request('{self.short_name}')"


class Response(BasicStructure):
    def __init__(self, id, short_name, parameters, long_name=None, response_type=None, description=None):
        super().__init__(id, short_name, parameters,
                         long_name=long_name, description=description)
        self.response_type = "POS-RESPONSE" if response_type == "POS-RESPONSE" else "NEG-RESPONSE"

    def encode(self, coded_request: bytearray = None, **params):
        logger.info(f"Compose response message to the request {coded_request}")
        if coded_request is not None:
            # Extract MATCHING-REQUEST-PARAMs from the coded request
            for param in self.parameters:
                if param.parameter_type == "MATCHING-REQUEST-PARAM":
                    logger.info(
                        f"set matching request param value {param.short_name}")
                    byte_pos = param.request_byte_position
                    byte_length = param.byte_length

                    val = coded_request[byte_pos:byte_pos+byte_length]
                    params[param.short_name] = val

        return super().encode(coded_request=coded_request, **params)

    def __repr__(self) -> str:
        return f"Response('{self.short_name}')"

    def __str__(self) -> str:
        return f"Response('{self.short_name}')"


def read_structure_from_odx(et_element) -> Union[Structure, Request, Response, None]:
    id = et_element.get("ID")
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.find("LONG-NAME")
    if long_name is not None:
        long_name = long_name.text
    description = read_description_from_odx(et_element.find("DESC"))
    parameters = [read_parameter_from_odx(et_parameter)
                  for et_parameter in et_element.iterfind("PARAMS/PARAM")]

    res: Union[Structure, Request, Response, None]
    if et_element.tag == "REQUEST":
        res = Request(
            id,
            short_name,
            parameters=parameters,
            long_name=long_name,
            description=description
        )
    elif et_element.tag in ["POS-RESPONSE", "NEG-RESPONSE"]:
        res = Response(
            id,
            short_name,
            response_type=et_element.tag,
            parameters=parameters,
            long_name=long_name,
            description=description
        )
    elif et_element.tag == "STRUCTURE":
        byte_size = int(et_element.find(
            "BYTE-SIZE").text) if et_element.find("BYTE-SIZE") is not None else None
        res = Structure(
            id,
            short_name,
            parameters=parameters,
            byte_size=byte_size,
            long_name=long_name,
            description=description
        )
    else:
        res = None
        logger.critical(
            f"Did not recognize structure {et_element.tag} {short_name}")
    return res



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from collections import OrderedDict
from dataclasses import dataclass
from typing import Optional, Dict, Any, List

from .utils import make_ref
from .structures import BasicStructure
from .dataobjectproperty import DopBase, DataObjectProperty
from .decodestate import DecodeState
from .encodestate import EncodeState
from .exceptions import DecodeError, EncodeError
from .globals import logger


@dataclass
class MultiplexerCase:
    """This class represents a Case which represents multiple options in a Multiplexer."""

    short_name: str
    long_name: str
    structure_ref: str
    lower_limit: str
    upper_limit: str

    def __post_init__(self):
        self._structure: Optional[BasicStructure] = None

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        self._structure = id_lookup.get(self.structure_ref)
        if self._structure is None:
            logger.warning(
                f"STRUCTURE-REF '{self.structure_ref}' could not be resolved."
            )

    def __repr__(self) -> str:
        return (
            f"MultiplexerCase('{self.short_name}', "
            + ", ".join(
                [
                    f"lower_limit='{self.lower_limit}'",
                    f"upper_limit='{self.upper_limit}'",
                    f"structure_ref='{self.structure_ref}'",
                ]
            )
            + ")"
        )


@dataclass
class MultiplexerDefaultCase:
    """This class represents a Default Case, which is selected when there are no cases defined in the Multiplexer."""

    short_name: str
    long_name: str
    structure_ref: Optional[str] = None

    def __post_init__(self):
        self._structure: Optional[BasicStructure] = None

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        if self.structure_ref is not None:
            self._structure = id_lookup.get(self.structure_ref)
            if self._structure is None:
                logger.warning(
                    f"STRUCTURE-REF '{self.structure_ref}' could not be resolved in DEFAULT-CASE."
                )

    def __repr__(self) -> str:
        return (
            f"MultiplexerDefaultCase('{self.short_name}', "
            + ", ".join(
                [
                    f"structure_ref='{self.structure_ref}'",
                ]
            )
            + ")"
        )


@dataclass
class MultiplexerSwitchKey:
    """This class represents a Switch Key, which is used to select one of the cases defined in the Multiplexer."""

    byte_position: int
    bit_position: int
    dop_ref: str

    def __post_init__(self):
        self._dop: DataObjectProperty = None # type: ignore

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        dop = id_lookup.get(self.dop_ref)
        if isinstance(dop, DataObjectProperty):
            self._dop = dop
        else:
            logger.warning(
                f"DATA-OBJECT-PROP-REF '{self.dop_ref}' could not be resolved in SWITCH-KEY."
            )

    def __repr__(self) -> str:
        return (
            f"MultiplexerSwitchKey("
            + ", ".join(
                [
                    f"byte_position='{self.byte_position}'",
                    f"bit_position='{self.bit_position}'",
                    f"dop_ref='{self.dop_ref}'",
                ]
            )
            + ")"
        )


@dataclass
class Multiplexer(DopBase):
    """This class represents a Multiplexer (MUX) which are used to interpret data stream depending on the value
    of a switch-key (similar to switch-case statements in programming languages like C or Java)."""

    id: str
    short_name: str
    long_name: str
    byte_position: int
    switch_key: MultiplexerSwitchKey
    default_case: Optional[MultiplexerDefaultCase] = None
    cases: Optional[List[MultiplexerCase]] = None

    @property
    def bit_length(self):
        return None

    def _get_case_limits(self, case: MultiplexerCase):
        key_type = self.switch_key._dop.physical_type.base_data_type
        lower_limit = key_type.make_from(case.lower_limit)
        upper_limit = key_type.make_from(case.upper_limit)
        return lower_limit, upper_limit

    def convert_physical_to_bytes(
            self,
            physical_value,
            encode_state: EncodeState,
            bit_position) -> bytes:

        if bit_position != 0:
            raise EncodeError(
                "Multiplexer must be aligned, i.e. bit_position=0, but "
                f"{self.short_name} was passed the bit position {bit_position}")

        if not isinstance(physical_value, dict) or len(physical_value) != 1:
            raise EncodeError("""Multiplexer should be defined as a dict
            with only one key equal to the desired case""")

        case_name, case_value = next(iter(physical_value.items()))
        key_pos = self.switch_key.byte_position
        case_pos = self.byte_position

        for case in (self.cases or []):
            if case.short_name == case_name:
                if case._structure:
                    case_bytes = case._structure.convert_physical_to_bytes(
                        case_value, encode_state, 0)
                else:
                    case_bytes = bytes()
                
                key_value, _ = self._get_case_limits(case)
                key_bytes = self.switch_key._dop.convert_physical_to_bytes(
                    key_value, encode_state, self.switch_key.bit_position)

                mux_len = max(
                    len(key_bytes) + key_pos,
                    len(case_bytes) + case_pos
                )
                mux_bytes = bytearray(mux_len)
                mux_bytes[key_pos:key_pos + len(key_bytes)] = key_bytes
                mux_bytes[case_pos:case_pos + len(case_bytes)] = case_bytes

                return bytes(mux_bytes)

        raise EncodeError(
            f"The case {case_name} is not found in Multiplexer {self.short_name}")

    def convert_bytes_to_physical(
            self,
            decode_state: DecodeState,
            bit_position: int = 0):

        if bit_position != 0:
            raise DecodeError(
                "Multiplexer must be aligned, i.e. bit_position=0, but "
                f"{self.short_name} was passed the bit position {bit_position}")
        byte_code = decode_state.coded_message[decode_state.next_byte_position:]
        key_decode_state = DecodeState(
            coded_message=byte_code[self.switch_key.byte_position:],
            parameter_value_pairs=[],
            next_byte_position=0
        )
        key_value, key_next_byte = self.switch_key._dop.convert_bytes_to_physical(
            key_decode_state, bit_position=self.switch_key.bit_position)

        case_decode_state = DecodeState(
            coded_message=byte_code[self.byte_position:],
            parameter_value_pairs=[],
            next_byte_position=0
        )
        case_found = False
        case_next_byte = 0
        case_value = None
        for case in (self.cases or []):
            lower, upper = self._get_case_limits(case)
            if lower <= key_value and key_value <= upper:
                case_found = True
                if case._structure:
                    case_value, case_next_byte = case._structure.convert_bytes_to_physical(
                        case_decode_state)
                break

        if not case_found and self.default_case is not None:
            case_found = True
            if self.default_case._structure:
                case_value, case_next_byte = self.default_case._structure.convert_bytes_to_physical(
                    case_decode_state)

        if not case_found:
            raise DecodeError(
                f"Failed to find a matching case in {self.short_name} for value {key_value}")

        mux_value = OrderedDict({case.short_name: case_value})
        mux_next_byte = decode_state.next_byte_position + \
            max(
                key_next_byte + self.switch_key.byte_position,
                case_next_byte + self.byte_position
            )
        return mux_value, mux_next_byte

    def _build_id_lookup(self):
        id_lookup = {}
        id_lookup.update({self.id: self})
        return id_lookup

    def _resolve_references(self, id_lookup: Dict[str, Any]) -> None:
        self.switch_key._resolve_references(id_lookup)
        if self.default_case is not None:
            self.default_case._resolve_references(id_lookup)

        for case in (self.cases or []):
            case._resolve_references(id_lookup)

    def __repr__(self) -> str:
        return (
            f"Multiplexer('{self.short_name}', "
            + ", ".join(
                [
                    f"id='{self.id}'",
                    f"byte_position='{self.byte_position}'",
                    f"switch_key='{self.switch_key}'",
                    f"default_case='{self.default_case}'",
                    f"cases='{self.cases}'",
                ]
            )
            + ")"
        )


def read_switch_key_from_odx(et_element):
    """Reads a Switch Key for a Multiplexer."""
    byte_position = (
        int(et_element.find("BYTE-POSITION").text)
        if et_element.find("BYTE-POSITION") is not None
        else None
    )
    bit_position = (
        int(et_element.find("BIT-POSITION").text)
        if et_element.find("BIT-POSITION") is not None
        else 0
    )
    dop_ref = make_ref(et_element.find("DATA-OBJECT-PROP-REF"))

    return MultiplexerSwitchKey(
        byte_position=byte_position,
        bit_position=bit_position,
        dop_ref=dop_ref,
    )


def read_default_case_from_odx(et_element):
    """Reads a Default Case for a Multiplexer."""
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.find("LONG-NAME")
    if long_name is not None:
        long_name = long_name.text

    structure_ref = make_ref(et_element.find("STRUCTURE-REF"))

    return MultiplexerDefaultCase(
        short_name=short_name,
        long_name=long_name,
        structure_ref=structure_ref,
    )


def read_case_from_odx(et_element):
    """Reads a Case for a Multiplexer."""
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.find("LONG-NAME")
    if long_name is not None:
        long_name = long_name.text
    structure_ref = make_ref(et_element.find("STRUCTURE-REF"))
    lower_limit = et_element.find("LOWER-LIMIT").text
    upper_limit = et_element.find("UPPER-LIMIT").text

    return MultiplexerCase(
        short_name=short_name,
        long_name=long_name,
        structure_ref=structure_ref,
        lower_limit=lower_limit,
        upper_limit=upper_limit,
    )


def read_mux_from_odx(et_element):
    """Reads a Multiplexer from Diag Layer."""
    id = et_element.get("ID")
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.find("LONG-NAME")
    if long_name is not None:
        long_name = long_name.text
    byte_position = (
        int(et_element.find("BYTE-POSITION").text)
        if et_element.find("BYTE-POSITION") is not None
        else None
    )
    switch_key = read_switch_key_from_odx(et_element.find("SWITCH-KEY"))

    default_case = None
    if et_element.find("DEFAULT-CASE") is not None:
        default_case = read_default_case_from_odx(
            et_element.find("DEFAULT-CASE"))

    cases = []
    if et_element.find("CASES") is not None:
        cases = [
            read_case_from_odx(el) for el in et_element.find("CASES").iterfind("CASE")
        ]

    logger.debug("Parsing MUX " + short_name)

    mux = Multiplexer(
        id=id,
        short_name=short_name,
        long_name=long_name,
        byte_position=byte_position,
        switch_key=switch_key,
        default_case=default_case,
        cases=cases,
    )

    return mux



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


import abc
from typing import Optional
import warnings

from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..exceptions import OdxWarning
from ..globals import logger


class Parameter(abc.ABC):
    def __init__(self,
                 short_name,
                 parameter_type,
                 long_name=None,
                 byte_position=None,
                 bit_position=0,
                 semantic=None,
                 description=None):
        self.short_name: str = short_name
        self.long_name: Optional[str] = long_name
        self.byte_position: Optional[int] = byte_position
        self.bit_position: int = bit_position
        self.parameter_type: str = parameter_type
        self.semantic: Optional[str] = semantic
        self.description: Optional[str] = description

    @property
    def bit_length(self) -> Optional[int]:
        return None

    @abc.abstractmethod
    def is_required(self):
        pass

    @abc.abstractmethod
    def is_optional(self):
        pass

    @abc.abstractmethod
    def get_coded_value(self):
        pass

    @abc.abstractmethod
    def get_coded_value_as_bytes(self, encode_state: EncodeState) -> bytes:
        """Get the coded value of the parameter given the encode state.
        Note that this method is called by `encode_into_pdu`.
        """
        pass

    @abc.abstractmethod
    def decode_from_pdu(self, decode_state: DecodeState):
        """Decode the parameter value from the coded message.

        If the paramter does have a byte position property, the coded bytes the parameter covers are extracted
        at this byte position and the function parameter `default_byte_position` is ignored.

        If the parameter does not have a byte position and a byte position is passed,
        the bytes are extracted at the byte position given by the argument `default_byte_position`.

        If the parameter does not have a byte position and the argument `default_byte_position` is None,
        this function throws a `DecodeError`.

        Parameters
        ----------
        decode_state : DecodeState
            The decoding state containing
            * the byte message to be decoded
            * the parameter values that are already decoded
            * the next byte position that is used iff the parameter does not specify a byte position

        Returns
        -------
        ParameterValuePair | List[ParameterValuePair]
            the decoded parameter value (the type is defined by the DOP)
        int
            the next byte position after the extracted parameter
        """
        pass

    def encode_into_pdu(self, encode_state: EncodeState) -> bytearray:
        """Insert the encoded value of a parameter into the coded RPC.

        If the byte position of the parameter is not defined,
        the byte code is appended to the coded RPC.

        Technical note for subclasses:
        The default implementation tries to compute the coded value
        via `self.get_coded_value_as_bytes(encoded_state)` and inserts it into
        the PDU. Thus it suffices to overwrite `get_coded_value_as_bytes()`.

        Parameters:
        ----------
        encode_state: EncodeState, i.e. a named tuple with attributes
            * coded_message: bytes, the message encoded so far
            * parameter_values: List[ParameterValuePairs]
            * triggering_coded_request: bytes

        Returns:
        -------
        bytearray
            the updated, coded message after encoding the parameter into it
        """
        byte_value = self.get_coded_value_as_bytes(encode_state)

        old_rpc = encode_state.coded_message
        new_rpc = bytearray(old_rpc)

        if self.byte_position is not None:
            byte_position = self.byte_position
        else:
            byte_position = len(old_rpc)

        min_length = byte_position + len(byte_value)
        if len(old_rpc) < min_length:
            # Make byte code longer if necessary
            new_rpc += bytearray([0] * (min_length - len(old_rpc)))
        for byte_idx_val, byte_idx_rpc in enumerate(
                range(byte_position, byte_position + len(byte_value))):
            # insert byte value
            if new_rpc[byte_idx_rpc] & byte_value[byte_idx_val] != 0:
                warnings.warn(
                    f"Parameter {self.short_name} overlaps with another parameter (bytes are already set)",
                    OdxWarning)
            new_rpc[byte_idx_rpc] |= byte_value[byte_idx_val]

        logger.debug(f"Param {self.short_name} inserts"
                     f" 0x{byte_value.hex()} at byte pos {byte_position}")
        return new_rpc

    def _as_dict(self):
        """
        Mostly for pretty printing purposes (specifically not for reconstructing the object)
        """
        d = {
            'short_name': self.short_name,
            'type': self.parameter_type,
            'semantic': self.semantic
        }
        if self.byte_position is not None:
            d['byte_position'] = self.byte_position
        if self.bit_position:
            d['bit_position'] = self.bit_position
        return d

    def __repr__(self):
        repr_str = f"Parameter(parameter_type='{self.parameter_type}', short_name='{self.short_name}'"
        if self.long_name is not None:
            repr_str += f", long_name='{self.long_name}'"
        if self.byte_position is not None:
            repr_str += f", byte_position='{self.byte_position}'"
        if self.bit_position:
            repr_str += f", bit_position='{self.bit_position}'"
        if self.semantic is not None:
            repr_str += f", semantic='{self.semantic}'"
        if self.description is not None:
            repr_str += f", description='{' '.join(self.description.split())}'"
        return repr_str + ")"

    def __str__(self):
        # create list of all parameters. 'short_name' ought to be
        # first, so it needs special treatment...
        param_descs = [f"short_name='{self.short_name}'"]
        for (key, val) in self._as_dict().items():
            if key == "short_name":
                continue
            elif isinstance(val, str):
                param_descs.append(f"{key}='{val}'")
            else:
                param_descs.append(f"{key}={val}")

        return f"Parameter({', '.join(param_descs)})"



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass

from odxtools.parameters import read_parameter_from_odx

from odxtools.utils import read_description_from_odx

from odxtools.structures import BasicStructure

from .globals import logger


@dataclass
class EnvironmentData(BasicStructure):
    """This class represents Environment Data that describes the circumstances in which the error occurred."""

    def __init__(
        self,
        id,
        short_name,
        parameters,
        long_name=None,
        description=None,
    ):
        super().__init__(
            id, short_name, parameters, long_name=long_name, description=description
        )

    def _build_id_lookup(self):
        id_lookup = {}
        id_lookup.update({self.id: self})
        return id_lookup

    def __repr__(self) -> str:
        return (
            f"EnvironmentData('{self.short_name}', "
            + ", ".join([f"id='{self.id}'", f"parameters='{self.parameters}'"])
            + ")"
        )


def read_env_data_from_odx(et_element):
    """Reads Environment Data from Diag Layer."""
    id = et_element.get("ID")
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.find("LONG-NAME")
    if long_name is not None:
        long_name = long_name.text
    description = read_description_from_odx(et_element.find("DESC"))
    parameters = [
        read_parameter_from_odx(et_parameter)
        for et_parameter in et_element.iterfind("PARAMS/PARAM")
    ]
    logger.debug("Parsing ENV-DATA " + short_name)

    env_data = EnvironmentData(
        id,
        short_name,
        parameters=parameters,
        long_name=long_name,
        description=description,
    )

    return env_data



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import abc
import math
from typing import Any, Union

from .utils import make_ref
from .odxtypes import DataType
from .exceptions import DecodeError, EncodeError
from .globals import xsi, logger
from .decodestate import DecodeState
from .encodestate import EncodeState

import bitstruct

ODX_TYPE_TO_FORMAT_LETTER = {
    DataType.A_INT32: 's',
    DataType.A_UINT32: 'u',
    DataType.A_FLOAT32: 'f',
    DataType.A_FLOAT64: 'f',
    DataType.A_BYTEFIELD: 'r',
    DataType.A_UNICODE2STRING: 'r',  # UTF-16 strings must be converted explicitly
    DataType.A_ASCIISTRING: 't',
    DataType.A_UTF8STRING: 't'
}


class DiagCodedType(abc.ABC):
    def __init__(self,
                 base_data_type: Union[str, DataType],
                 dct_type: str,
                 base_type_encoding=None,
                 is_highlow_byte_order: bool = True):
        self.base_data_type = DataType(base_data_type)
        self.dct_type = dct_type
        self.base_type_encoding = base_type_encoding
        self.is_highlow_byte_order = is_highlow_byte_order

    def _extract_internal(self,
                          coded_message,
                          byte_position,
                          bit_position,
                          bit_length,
                          base_data_type: DataType,
                          is_highlow_byte_order,
                          bit_mask=None):
        """Extract the internal value.

        Helper method for `DiagCodedType.convert_bytes_to_internal`.
        """
        # If the bit length is zero, return "empty" values of each type
        if bit_length == 0:
            return base_data_type.as_python_type()(), byte_position

        byte_length = (bit_length + bit_position + 7) // 8
        if byte_position + byte_length > len(coded_message):
            raise DecodeError(
                f"Expected a longer message."
            )
        next_byte_position = byte_position + byte_length
        extracted_bytes = coded_message[byte_position:next_byte_position]

        # TODO: Apply bit mask, etc.
        if bit_mask is not None:
            raise NotImplementedError(
                f"Don't know how to handle bit_mask={bit_mask}.")

        # Apply byteorder
        if not is_highlow_byte_order and base_data_type not in [DataType.A_UNICODE2STRING, DataType.A_BYTEFIELD, DataType.A_ASCIISTRING, DataType.A_UTF8STRING]:
            extracted_bytes = extracted_bytes[::-1]

        format_letter = ODX_TYPE_TO_FORMAT_LETTER[base_data_type]
        padding = 8 * byte_length - (bit_length + bit_position)
        internal_value = bitstruct.unpack_from(
            f"{format_letter}{bit_length}", extracted_bytes, offset=padding)[0]

        if base_data_type == DataType.A_UNICODE2STRING:
            # Convert bytes to string with utf-16 decoding
            if is_highlow_byte_order:
                internal_value = internal_value.decode("utf-16-be")
            else:
                internal_value = internal_value.decode("utf-16-le")

        return internal_value, next_byte_position

    def _to_bytes(self, internal_value, bit_position, bit_length, base_data_type, is_highlow_byte_order, bit_mask=None):
        """Convert the internal_value to bytes."""
        # Check that bytes and strings actually fit into the bit length
        if base_data_type in [DataType.A_BYTEFIELD] and 8 * len(internal_value) > bit_length:
            raise EncodeError(f"The bytefield {internal_value.hex()} is too large."
                              f" The maximum byte length is {bit_length//8}.")
        if base_data_type in [DataType.A_ASCIISTRING, DataType.A_UTF8STRING] and 8 * len(internal_value) > bit_length:
            raise EncodeError(f"The string {repr(internal_value)} is too large."
                              f" The maximum number of characters is {bit_length//8}.")
        if base_data_type in [DataType.A_UNICODE2STRING] and 16 * len(internal_value) > bit_length:
            raise EncodeError(f"The string {repr(internal_value)} is too large."
                              f" The maximum number of characters is {bit_length//16}.")

        # If the bit length is zero, return empty bytes
        if bit_length == 0:
            if base_data_type in [DataType.A_INT32, DataType.A_UINT32, DataType.A_FLOAT32, DataType.A_FLOAT64] and base_data_type != 0:
                raise EncodeError(
                    f"The number {repr(internal_value)} cannot be encoded into {bit_length} bits.")
            return bytes()

        char = ODX_TYPE_TO_FORMAT_LETTER[base_data_type]

        # The coded byte is divided into (0..0)(value)(0..0) with bit lengths (left_pad)(bit_length)(bit_position)
        offset = (8 - ((bit_length + bit_position) % 8)) % 8
        assert 0 <= offset and offset < 8 and (
            offset + bit_length + bit_position) % 8 == 0, f"Computational mistake, offset={offset}"
        left_pad = f'p{offset}' if offset > 0 else ''

        # Convert string to bytes with utf-16 encoding
        if base_data_type == DataType.A_UNICODE2STRING:
            if is_highlow_byte_order:
                internal_value = internal_value.encode("utf-16-be")
            else:
                internal_value = internal_value.encode("utf-16-le")

        code = bitstruct.pack(f'{left_pad}{char}{bit_length}', internal_value)

        if not is_highlow_byte_order and base_data_type not in [DataType.A_UNICODE2STRING, DataType.A_BYTEFIELD, DataType.A_ASCIISTRING, DataType.A_UTF8STRING]:
            code = code[::-1]

        # TODO: Apply bit mask.
        if bit_mask is not None:
            raise NotImplementedError(
                f"Don't know how to handle bit_mask={bit_mask}."
            )

        return code

    def _minimal_byte_length_of(self, internal_value: Union[bytes, str]) -> int:
        """Helper method to get the minimal byte length.
        (needed for LeadingLength- and MinMaxLengthType)
        """
        # A_BYTEFIELD, A_ASCIISTRING, A_UNICODE2STRING, A_UTF8STRING
        if self.base_data_type == DataType.A_BYTEFIELD:
            byte_length = len(internal_value)
        elif self.base_data_type in [DataType.A_ASCIISTRING, DataType.A_UTF8STRING]:
            assert isinstance(internal_value, str)
            # TODO: Handle different encodings
            byte_length = len(bytes(internal_value, "utf-8"))
        elif self.base_data_type == DataType.A_UNICODE2STRING:
            assert isinstance(internal_value, str)
            byte_length = len(bytes(internal_value, "utf-16-le"))
            assert byte_length % 2 == 0, (f"The bit length of A_UNICODE2STRING must"
                                          f" be a multiple of 16 but is {8*byte_length}")
        return byte_length

    @abc.abstractmethod
    def convert_internal_to_bytes(self,
                                  internal_value: Any,
                                  encode_state: EncodeState,
                                  bit_position: int) \
                                  -> Union[bytes, bytearray]:
        """Encode the internal value.

        Parameters
        ----------
        internal_value : python type corresponding to self.base_data_type
            the value to be encoded
        bit_position : int 

        length_keys : Dict[str, int]
            mapping from ID (of the length key) to bit length
            (only needed for ParamLengthInfoType) 
        """
        pass

    @abc.abstractmethod
    def convert_bytes_to_internal(self, decode_state: DecodeState, bit_position: int = 0) -> Any:
        """Decode the parameter value from the coded message.

        Parameters
        ----------
        decode_state : DecodeState
            The decoding state

        Returns
        -------
        str or int or bytes or bytearray or dict
            the decoded parameter value
        int
            the next byte position after the extracted parameter
        """
        pass


class LeadingLengthInfoType(DiagCodedType):
    def __init__(self,
                 base_data_type,
                 bit_length,
                 base_type_encoding=None,
                 is_highlow_byte_order=True):
        super().__init__(base_data_type,
                         dct_type="LEADING-LENGTH-INFO-TYPE",
                         base_type_encoding=base_type_encoding,
                         is_highlow_byte_order=is_highlow_byte_order)
        self.bit_length = bit_length
        assert self.bit_length > 0, "A Leading length info type with bit length == 0 does not make sense."
        assert self.base_data_type in [DataType.A_BYTEFIELD, DataType.A_ASCIISTRING, DataType.A_UNICODE2STRING, DataType.A_UTF8STRING], (
            f"A leading length info type cannot have the base data type {self.base_data_type}."
        )

    def convert_internal_to_bytes(self,
                                  internal_value: Any,
                                  encode_state: EncodeState,
                                  bit_position: int) \
                                  -> Union[bytes, bytearray]:

        byte_length = self._minimal_byte_length_of(internal_value)

        length_byte = self._to_bytes(byte_length,
                                     bit_position=bit_position,
                                     bit_length=self.bit_length,
                                     base_data_type=DataType.A_UINT32,
                                     is_highlow_byte_order=self.is_highlow_byte_order)

        value_byte = self._to_bytes(internal_value,
                                    bit_position=0,
                                    bit_length=8*byte_length,
                                    base_data_type=self.base_data_type,
                                    is_highlow_byte_order=self.is_highlow_byte_order)

        return length_byte + value_byte

    def convert_bytes_to_internal(self, decode_state: DecodeState, bit_position: int = 0):
        coded_message = decode_state.coded_message

        # Extract length of the parameter value
        byte_length, byte_position = self._extract_internal(coded_message=coded_message,
                                                            byte_position=decode_state.next_byte_position,
                                                            bit_position=bit_position,
                                                            bit_length=self.bit_length,
                                                            base_data_type=DataType.A_UINT32,  # length is an integer
                                                            is_highlow_byte_order=self.is_highlow_byte_order)

        # Extract actual value
        # TODO: The returned value is None if the byte_length is 0. Maybe change it
        #       to some default value like an empty bytearray() or 0?
        value, next_byte_position = self._extract_internal(coded_message=coded_message,
                                                           byte_position=byte_position,
                                                           bit_position=0,
                                                           bit_length=8*byte_length,
                                                           base_data_type=self.base_data_type,
                                                           is_highlow_byte_order=self.is_highlow_byte_order)

        return value, next_byte_position

    def __repr__(self) -> str:
        repr_str = f"LeadingLengthInfoType(base_data_type='{self.base_data_type}', bit_length={self.bit_length}"
        if self.base_type_encoding is not None:
            repr_str += f", base_type_encoding={self.base_type_encoding}"
        if not self.is_highlow_byte_order:
            repr_str += f", is_highlow_byte_order={self.is_highlow_byte_order}"
        return repr_str + ")"

    def __str__(self) -> str:
        return self.__repr__()


class MinMaxLengthType(DiagCodedType):
    def __init__(self,
                 base_data_type,
                 min_length,
                 termination,
                 max_length=None,
                 base_type_encoding=None,
                 is_highlow_byte_order=True):
        super().__init__(base_data_type,
                         dct_type="MIN-MAX-LENGTH-TYPE",
                         base_type_encoding=base_type_encoding,
                         is_highlow_byte_order=is_highlow_byte_order)
        assert max_length is None or min_length <= max_length
        self.min_length = min_length
        self.max_length = max_length
        self.termination = termination

        assert self.base_data_type in [DataType.A_BYTEFIELD, DataType.A_ASCIISTRING, DataType.A_UNICODE2STRING, DataType.A_UTF8STRING], (
            f"A min-max length type cannot have the base data type {self.base_data_type}."
        )
        assert self.termination in ["ZERO", "HEX-FF", "END-OF-PDU"], (
            f"A min-max length type cannot have the termination {self.termination}"
        )

    def __termination_character(self):
        """Returns the termination character or None if it isn't defined."""
        # The termination character is actually not specified by ASAM
        # for A_BYTEFIELD but I assume it is only one byte.
        termination_char = None
        if self.termination == "ZERO":
            if self.base_data_type not in [DataType.A_UNICODE2STRING]:
                termination_char = bytes([0x0])
            else:
                termination_char = bytes([0x0, 0x0])
        elif self.termination == "HEX-FF":
            if self.base_data_type not in [DataType.A_UNICODE2STRING]:
                termination_char = bytes([0xff])
            else:
                termination_char = bytes([0xff, 0xff])
        return termination_char

    def convert_internal_to_bytes(self, internal_value, encode_state: EncodeState, bit_position: int) -> bytes:
        byte_length = self._minimal_byte_length_of(internal_value)

        # The coded value must have at least length min_length
        if byte_length < self.min_length:
            raise EncodeError(f"The internal value {internal_value} is only {byte_length} bytes"
                              f" long but the min length is {self.min_length}")
        # The coded value must not have a length greater than max_length
        if self.max_length and byte_length > self.max_length:
            raise EncodeError(f"The internal value {internal_value} requires {byte_length}"
                              f" bytes, but the max length is {self.max_length}")

        value_byte = self._to_bytes(internal_value,
                                    bit_position=0,
                                    bit_length=8*byte_length,
                                    base_data_type=self.base_data_type,
                                    is_highlow_byte_order=self.is_highlow_byte_order)

        if encode_state.is_end_of_pdu or byte_length == self.max_length:
            # All termination types may be ended by the PDU
            return value_byte
        else:
            termination_char = self.__termination_character()
            if self.termination == 'END-OF-PDU':
                termination_char = bytes()
            assert termination_char is not None, \
                (f"MinMaxLengthType with termination {self.termination}"
                 f"(min: {self.min_length}, max: {self.max_length}) failed encoding {internal_value}")
            return value_byte + termination_char

    def convert_bytes_to_internal(self, decode_state: DecodeState, bit_position: int = 0):
        if decode_state.next_byte_position + self.min_length > len(decode_state.coded_message):
            raise DecodeError("The PDU ended before min length was reached.")

        coded_message = decode_state.coded_message
        byte_position = decode_state.next_byte_position
        termination_char = self.__termination_character()

        # If no termination char is found, this is the next byte after the parameter.
        max_termination_byte = len(coded_message)
        if self.max_length is not None:
            max_termination_byte = min(max_termination_byte, byte_position + self.max_length)

        if self.termination != "END-OF-PDU":
            # The parameter either ends after max length, at the end of the PDU
            # or if a termination character is found.
            char_length = len(termination_char)  # either 1 or 2

            termination_byte = byte_position + self.min_length
            found_char = False
            # Search the termination character
            while termination_byte < max_termination_byte and not found_char:
                found_char = coded_message[termination_byte: termination_byte +
                                           char_length] == termination_char
                if not found_char:
                    termination_byte += char_length

            byte_length = termination_byte - byte_position

            # Extract the value
            value, byte = self._extract_internal(decode_state.coded_message,
                                                 byte_position=byte_position,
                                                 bit_position=bit_position,
                                                 bit_length=8*byte_length,
                                                 base_data_type=self.base_data_type,
                                                 is_highlow_byte_order=self.is_highlow_byte_order)
            assert byte == termination_byte

            # next byte starts after the termination character
            next_byte_position = byte+char_length if found_char else byte
            return value, next_byte_position
        else:
            # If termination == "END-OF-PDU", the parameter ends after max_length
            # or at the end of the PDU.
            byte_length = max_termination_byte - byte_position

            value, byte = self._extract_internal(decode_state.coded_message,
                                                 byte_position=byte_position,
                                                 bit_position=bit_position,
                                                 bit_length=8*byte_length,
                                                 base_data_type=self.base_data_type,
                                                 is_highlow_byte_order=self.is_highlow_byte_order)
            return value, byte

    def __repr__(self) -> str:
        repr_str = f"MinMaxLengthType(base_data_type='{self.base_data_type}', min_length={self.min_length}"
        if self.max_length is not None:
            repr_str += f", base_type_encoding={self.max_length}"
        if self.termination is not None:
            repr_str += f", termination={self.termination}"
        if self.base_type_encoding is not None:
            repr_str += f", base_type_encoding={self.base_type_encoding}"
        if not self.is_highlow_byte_order:
            repr_str += f", is_highlow_byte_order={self.is_highlow_byte_order}"
        return repr_str + ")"

    def __str__(self) -> str:
        return self.__repr__()


class ParamLengthInfoType(DiagCodedType):
    def __init__(self,
                 base_data_type: Union[str, DataType],
                 length_key_ref: str,
                 base_type_encoding=None,
                 is_highlow_byte_order=True):
        super().__init__(base_data_type,
                         dct_type="PARAM-LENGTH-INFO-TYPE",
                         base_type_encoding=base_type_encoding,
                         is_highlow_byte_order=is_highlow_byte_order)
        self.length_key_ref = length_key_ref

    def convert_internal_to_bytes(self, internal_value, encode_state: EncodeState, bit_position: int) -> bytes:
        bit_length = encode_state.length_keys.get(self.length_key_ref, None)

        if bit_length is None:
            if self.base_data_type in [DataType.A_BYTEFIELD, DataType.A_ASCIISTRING, DataType.A_UTF8STRING]:
                bit_length = 8 * len(internal_value)
            if self.base_data_type in [DataType.A_UNICODE2STRING]:
                bit_length = 16 * len(internal_value)
        
            if self.base_data_type in [DataType.A_INT32, DataType.A_UINT32]:
                bit_length = int(internal_value).bit_length()
                if self.base_data_type == DataType.A_INT32:
                    bit_length += 1
                # Round up
                bit_length = math.ceil(bit_length / 8.0) * 8

        assert bit_length is not None
        encode_state.length_keys[self.length_key_ref] = bit_length

        return self._to_bytes(internal_value,
                              bit_position=bit_position,
                              bit_length=bit_length,
                              base_data_type=self.base_data_type,
                              is_highlow_byte_order=self.is_highlow_byte_order)

    def convert_bytes_to_internal(self, decode_state: DecodeState, bit_position: int = 0):
        # Find length key with matching ID.
        bit_length = None
        for parameter, value in decode_state.parameter_value_pairs:
            # if isinstance(param_value.parameter, LengthKeyParameter) would be prettier,
            # but leads to cyclic import...
            if parameter.parameter_type == "LENGTH-KEY" \
                    and parameter.id == self.length_key_ref:
                # The bit length of the parameter to be extracted is given by the length key.
                bit_length = value
                break

        assert value is not None, f"Did not find any length key with ID {self.length_key_ref}"

        # Extract the internal value and return.
        return self._extract_internal(decode_state.coded_message,
                                      decode_state.next_byte_position,
                                      bit_position,
                                      bit_length,
                                      self.base_data_type,
                                      self.is_highlow_byte_order)

    def __repr__(self) -> str:
        repr_str = f"ParamLengthInfoType(base_data_type='{self.base_data_type}', length_key_ref={self.length_key_ref}"
        if self.base_type_encoding is not None:
            repr_str += f", base_type_encoding={self.base_type_encoding}"
        if not self.is_highlow_byte_order:
            repr_str += f", is_highlow_byte_order={self.is_highlow_byte_order}"
        return repr_str + ")"

    def __str__(self) -> str:
        return self.__repr__()


class StandardLengthType(DiagCodedType):

    def __init__(self,
                 base_data_type: Union[str, DataType],
                 bit_length: int,
                 bit_mask=None,
                 condensed: bool = False,
                 base_type_encoding=None,
                 is_highlow_byte_order=True):
        super().__init__(base_data_type,
                         dct_type="STANDARD-LENGTH-TYPE",
                         base_type_encoding=base_type_encoding,
                         is_highlow_byte_order=is_highlow_byte_order)
        self.bit_length = bit_length
        self.bit_mask = bit_mask
        self.condensed = condensed

    def convert_internal_to_bytes(self, internal_value, encode_state: EncodeState, bit_position: int) -> bytes:
        return self._to_bytes(internal_value,
                              bit_position,
                              self.bit_length,
                              self.base_data_type,
                              is_highlow_byte_order=self.is_highlow_byte_order,
                              bit_mask=self.bit_mask)

    def convert_bytes_to_internal(self, decode_state: DecodeState, bit_position: int = 0):
        return self._extract_internal(decode_state.coded_message,
                                      decode_state.next_byte_position,
                                      bit_position,
                                      self.bit_length,
                                      self.base_data_type,
                                      self.is_highlow_byte_order,
                                      bit_mask=self.bit_mask)

    def __repr__(self) -> str:
        repr_str = f"StandardLengthType(base_data_type='{self.base_data_type}', bit_length={self.bit_length}"
        if self.bit_mask is not None:
            repr_str += f", bit_mask={self.bit_mask}"
        if self.condensed:
            repr_str += f", condensed={self.condensed}"
        if self.base_type_encoding is not None:
            repr_str += f", base_type_encoding={self.base_type_encoding}"
        if not self.is_highlow_byte_order:
            repr_str += f", is_highlow_byte_order={self.is_highlow_byte_order}"
        return repr_str + ")"

    def __str__(self) -> str:
        return self.__repr__()


def read_diag_coded_type_from_odx(et_element):
    base_type_encoding = et_element.get("BASE-TYPE-ENCODING")

    base_data_type = et_element.get("BASE-DATA-TYPE")
    assert base_data_type in ["A_INT32", "A_UINT32", "A_FLOAT32", "A_FLOAT64",
                              "A_ASCIISTRING", "A_UTF8STRING", "A_UNICODE2STRING", "A_BYTEFIELD"]

    if et_element.get("IS-HIGHLOW-BYTE-ORDER") is not None:
        assert et_element.get("IS-HIGHLOW-BYTE-ORDER") in ["true", "false"]
        is_highlow_byte_order = et_element.get(
            "IS-HIGHLOW-BYTE-ORDER") == "true"
        logger.debug(
            f"HIGH-LOW-BYTE-ORDER set to {is_highlow_byte_order}")
    else:
        is_highlow_byte_order = True

    dct_type = et_element.get(f"{xsi}type")
    bit_length = None
    if dct_type == "LEADING-LENGTH-INFO-TYPE":
        bit_length = int(et_element.find("BIT-LENGTH").text)
        return LeadingLengthInfoType(base_data_type,
                                     bit_length=bit_length,
                                     base_type_encoding=base_type_encoding,
                                     is_highlow_byte_order=is_highlow_byte_order)
    elif dct_type == "MIN-MAX-LENGTH-TYPE":
        min_length = int(et_element.find("MIN-LENGTH").text)
        max_length = None
        if et_element.find("MAX-LENGTH"):
            max_length = int(et_element.find("MAX-LENGTH").text)
        termination = et_element.get("TERMINATION")

        return MinMaxLengthType(base_data_type,
                                min_length=min_length,
                                max_length=max_length,
                                termination=termination,
                                base_type_encoding=base_type_encoding,
                                is_highlow_byte_order=is_highlow_byte_order)
    elif dct_type == "PARAM-LENGTH-INFO-TYPE":
        length_key_ref = make_ref(et_element.find("LENGTH-KEY-REF"))

        return ParamLengthInfoType(base_data_type,
                                   length_key_ref,
                                   base_type_encoding=base_type_encoding,
                                   is_highlow_byte_order=is_highlow_byte_order)
    elif dct_type == "STANDARD-LENGTH-TYPE":
        bit_length = int(et_element.find("BIT-LENGTH").text)
        bit_mask = None
        if et_element.find("BIT-MASK"):
            bit_mask = et_element.find("BIT-MASK").text
        condensed = et_element.get("CONDENSED") == "true"
        return StandardLengthType(base_data_type,
                                  bit_length,
                                  bit_mask=bit_mask,
                                  condensed=condensed,
                                  base_type_encoding=base_type_encoding,
                                  is_highlow_byte_order=is_highlow_byte_order)
    raise NotImplementedError(f"I do not know the diag-coded-type {dct_type}")



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass
from typing import List, Optional

from .utils import make_ref
from .globals import logger


@dataclass
class EnvironmentDataDescription:
    """This class represents Environment Data Description, which is a complex DOP
    that is used to define the interpretation of environment data."""

    id: str
    short_name: str
    long_name: str
    env_data_refs: List[str]
    param_snref: Optional[str] = None
    param_snpathref: Optional[str] = None

    def _build_id_lookup(self):
        id_lookup = {}
        id_lookup.update({self.id: self})
        return id_lookup

    def __repr__(self) -> str:
        return (
            f"EnvironmentDataDescription('{self.short_name}', "
            + ", ".join(
                [
                    f"id='{self.id}'",
                    f"param_snref='{self.param_snref}'",
                    f"param_snpathref='{self.param_snpathref}'",
                    f"env_data_refs='{self.env_data_refs}'",
                ]
            )
            + ")"
        )


def read_env_data_desc_from_odx(et_element):
    """Reads Environment Data Description from Diag Layer."""
    id = et_element.get("ID")
    short_name = et_element.find("SHORT-NAME").text
    long_name = et_element.find("LONG-NAME").text
    param_snref = None
    if et_element.find("PARAM-SNREF") is not None:
        param_snref = et_element.find("PARAM-SNREF").get("SHORT-NAME")
    param_snpathref = None
    if et_element.find("PARAM-SNPATHREF") is not None:
        param_snpathref = et_element.find("PARAM-SNPATHREF").get("SHORT-NAME-PATH")
    env_data_refs = []
    env_data_refs.extend([
        make_ref(env_data_ref)
        for env_data_ref in et_element.iterfind("ENV-DATA-REFS/ENV-DATA-REF")
    ])
    # ODX 2.0.0 says ENV-DATA-DESC could contain a list of ENV-DATAS
    env_data_refs.extend([
        env_data.get("ID")
        for env_data in et_element.iterfind("ENV-DATAS/ENV-DATA")
    ])
    logger.debug("Parsing ENV-DATA-DESC " + short_name)

    env_data_desc = EnvironmentDataDescription(
        id=id,
        short_name=short_name,
        long_name=long_name,
        param_snref=param_snref,
        param_snpathref=param_snpathref,
        env_data_refs=env_data_refs,
    )

    return env_data_desc
